DEBUG = false
INTEGRATED_PERIOD = 43
PLAYER_HEALTH = 175
PLAYER_SHIELD = 300
SEEKING = true
FRAGMENT = "triple health"
FX_BLOOM = ""



Triggers = {}
CollectionsUsed = {
   2, 3, 5, 8, 9, 10, 11,
   12, 13,14, 15, 16, 31
}
AllowedMonsters = {
   "minor compiler", "major compiler", "minor invisible compiler",
   "major invisible compiler", "minor defender", "major defender",
}
OverlayColors = {
   "green", "white", "red", "dark green",
   "cyan", "yellow", "dark red", "blue"
}

LF = "\n"

minStrLen = 256
maxStrLen = 512
lowByte = 32
highByte = 126

maxRpsInterval = 25
maxFaderInterval = 60
maxSoundInterval = 30
maxHudInterval = 5
maxHudColorInterval = 8
maxHudIconInterval = 30
maxPowerupInterval = 90
maxPowerupDuration = 30
maxWeaponSwitchInterval = 120
maxColorInterval = 3
maxTeamInterval = 4
maxCrosshairInterval = 8
maxCompassInterval = 10
maxItemInterval = 15
maxOldCoords = 30
maxOldCoordInterval = 30 * 30

minFogDepth = 3
maxFogDepth = 20
maxFogInterval = 30
fogColorInterval = 15
fogDepthInterval = 15
maxUwFogInterval = 30
uwFogColorInterval = 15
uwFogDepthInterval = 15

maxMonsterInterval = 30 * 7
monsterInterval = 30
maxMonsterSeed = 40

maxLightPeriod = 15

hudBufferSize = 512

function randomNumber(min, max)
   return Game.random(max - min + 1) + min
end

function randomRGB()
   return zeroPad(string.format("%X", randomNumber(0,255)))
end

function zeroPad(n)
   while string.len(n) < 2 do
      n = "0"..n
   end
   return n
end

function ballsackXY(polygon)
   local lowestX, lowestY, highestX, highestY = 33,33,-33,-33
   for e in polygon:endpoints() do
      if e.x < lowestX then lowestX = e.x end
      if e.y < lowestY then lowestY = e.y end
      if e.x > highestX then highestX = e.x end
      if e.y > highestY then highestY = e.y end
   end
   local tries = 0
   while (tries < 5) do
      tries = tries + 1
      local x = lowestX + Game.random(highestX * 1024 - lowestX * 1024) / 1024
      local y = lowestY + Game.random(highestY * 1024 - lowestY * 1024) / 1024
      if polygon:contains(x, y) then
	 return x, y
      end
   end
   return polygon.x, polygon.y
end

function randomString(min, max)
   if not min or not max then
      min = minStrLen
      max = maxStrLen
   end
   local str = ""
   local length = randomNumber(min, max)
   for i=1, length do
      str = str..string.char(randomNumber(lowByte, highByte))
   end
   return str
end

function randomChar()
   return string.char(randomNumber(lowByte, highByte))
end

function ticksTest(interval)
   return math.mod(Game.ticks, interval) == 0
end

function randomSound()
   return Sounds[randomNumber(1, #Sounds - 1)]
end

function randomFader()
   return FadeTypes[randomNumber(1, #FadeTypes - 1)]
end

function randomTerminalColor()
   return OverlayColors[randomNumber(1, #OverlayColors - 1)]
end

function randomPrintScreen(player)
   if ticksTest(player._rpsInterval) then
      player:print(randomString())
   end
end

function randomNoiseEmanations(player)
   if ticksTest(player._soundInterval) then
      player.monster:play_sound(randomSound())
   end
end

function randomScreenFades(player)
   if ticksTest(player._faderInterval) then
      player:fade_screen(randomFader())
   end
end

function randomHudColor(player)
   if ticksTest(player._hudInterval) then
      player.overlays[0].color = randomTerminalColor()
   end
end

function advanceHudString(player)
   if ticksTest(player._hudInterval) then
      player._hudString =
	 string.sub(player._hudString, 2)..randomChar()
      player.overlays[0].text = player._hudString
   end
end

function generateIcon()
   local nColors = randomNumber(1, 4)
   local icon = nColors
   for i=0, nColors - 1 do
      icon = icon..LF..string.char(58+i)
      local r, g, b = randomRGB(), randomRGB(), randomRGB()
      icon = icon..r..g..b
   end
   for i=1,16 do
      icon = icon..LF
      for i=1,16 do
	 icon = icon..string.char(58+randomNumber(0,nColors-1))
      end
   end
   return icon..LF
end

function randomHudIcon(player)
   if ticksTest(player._hudIconInterval) then
      player.overlays[0].icon = generateIcon()
   end
end

function randomizedPowerup(player)
   if ticksTest(player._powerupInterval) then
      local n = randomNumber(0, 3)
      local duration = randomNumber(0, maxPowerupDuration)
      if n == 0 then
	 player.extravision_duration = duration
      elseif n == 1 then
	 player.infravision_duration = duration
      elseif n == 2 then
	 player.invincibility_duration = duration
      elseif n == 3 then
	 player.invisibility_duration = duration
      end
   end
end

function randomWeaponSwitch(player)
   if ticksTest(player._weaponSwitchInterval) then
      player.weapons.active = not player.weapons.active
   end
end

function randomColor(player)
   player.color = randomNumber(0, 7)
end

function randomTeam(player)
   player.team = randomNumber(0,7)
end

function randomColors(player)
   randomColor(player)
   randomTeam(player)
end

function randomCrosshair(player)
   if ticksTest(player._crosshairInterval) then
      player.crosshairs.active = not player.crosshairs.active
   end
end

function randomCompass(player)
   if ticksTest(player._compassInterval) then
      player.compass:all_off()
      local n = randomNumber(0,1)
      if n == 1 then player.compass.ne = true end
      n = randomNumber(0,1)
      if n == 1 then player.compass.nw = true end
      n = randomNumber(0,1)
      if n == 1 then player.compass.se = true end
      n = randomNumber(0,1)
      if n == 1 then player.compass.sw = true end
   end
end

function spawnNewItem(player)
   player._item.object =
      Items.new(player.x, player.y, player.z + player._item.delta,
		player.polygon, player._item.type)
end

function randomItem(player)
   if ticksTest(player._itemInterval) then
      player._item.type = ItemTypes[randomNumber(1, #ItemTypes - 1)]
      while player._item.type == "alien weapon ammo" do
	 player._item.type = ItemTypes[randomNumber(1, #ItemTypes - 1)]
      end
   end
end

function randomizePlayerIntervals(player)
   player._rpsInterval = randomNumber(1, maxRpsInterval)
   player._faderInterval = randomNumber(1, maxFaderInterval)
   player._soundInterval = randomNumber(1, maxSoundInterval)
   player._hudInterval = randomNumber(1, maxHudInterval)
   player._hudColorInterval = randomNumber(1, maxHudColorInterval)
   player._hudIconInterval = randomNumber(1, maxHudIconInterval)
   player._powerupInterval = randomNumber(1, maxPowerupInterval)
   player._weaponSwitchInterval = randomNumber(1, maxWeaponSwitchInterval)
   player._colorInterval = randomNumber(1, maxColorInterval)
   player._teamInterval = randomNumber(1, maxTeamInterval)
   player._crosshairInterval = randomNumber(1, maxCrosshairInterval)
   player._compassInterval = randomNumber(1, maxCompassInterval)
   player._itemInterval = randomNumber(1, maxItemInterval)
   player._oldCoordInterval = randomNumber(1, maxOldCoordInterval)
end

function randomHudStuff(player)
   advanceHudString(player)
   randomHudColor(player)
   randomHudIcon(player)
end

function playerItemUpdate(player)
   local valid = false
   for i in Items() do
      if i == player._item.object then valid = true end
   end
   if player._item.object and valid then
      player._item.object:delete()
   end
   spawnNewItem(player)
end

function playerCoordsUpdate(player)
   local coords = {}
   coords.x, coords.y, coords.z, coords.polygon,
   coords.yaw, coords.pitch, coords.vx, coords.vy, coords.vz =
      player.monster.x, player.monster.y, player.monster.z,
      player.monster.polygon, player.yaw, player.pitch,
      player.external_velocity.x, player.external_velocity.y,
      player.external_velocity.z
   if #player._oldCoords >= maxOldCoords then
      table.remove(player._oldCoords, 1)
   end
   table.insert(player._oldCoords, coords)
end

function lagPlayer(player)
   if #player._oldCoords >=  maxOldCoords then
      if ticksTest(player._oldCoordInterval) then
	 local coords = table.remove(player._oldCoords, 1)
	 player:position(coords.x, coords.y, coords.z, coords.polygon)
	 player.pitch, player.yaw = coords.pitch, coords.yaw
	 player.external_velocity.x, player.external_velocity.y,
	 player.external_velocity.z =
	    coords.vx, coords.vy, coords.vz
	 player._oldCoords = {}
      end
   end
end

function terrorizePlayers()
   for p in Players() do
      randomizePlayerIntervals(p)
      playerCoordsUpdate(p)
      playerItemUpdate(p)
      lagPlayer(p)
      randomPrintScreen(p)
      randomNoiseEmanations(p)
      randomScreenFades(p)
      randomHudStuff(p)
      randomizedPowerup(p)
      randomWeaponSwitch(p)
      randomColors(p)
      randomCrosshair(p)
      randomCompass(p)
      randomItem(p)
   end
end

function randomOverFog()
   if ticksTest(fogColorInterval) then
      Level.fog.color.r = randomNumber(0, 255) / 255
      Level.fog.color.g = randomNumber(0, 255) / 255
      Level.fog.color.b = randomNumber(0, 255) / 255
   end
   if ticksTest(fogDepthInterval) then
      Level.fog.depth = randomNumber(minFogDepth, maxFogDepth)
   end
end

function randomUnderFog()
   if ticksTest(uwFogColorInterval) then
      Level.underwater_fog.color.r = randomNumber(0, 255) / 255
      Level.underwater_fog.color.g = randomNumber(0, 255) / 255
      Level.underwater_fog.color.b = randomNumber(0, 255) / 255
   end
   if ticksTest(uwFogDepthInterval) then
      Level.underwater_fog.depth = randomNumber(minFogDepth, maxFogDepth)
   end
end

function randomFoggyLandscapes()
   if ticksTest(fogLandscapesInterval) then
      Level.fog.affects_landscapes = not Level.fog.affects_landscapes
      Level.underwater_fog.affects_landscapes =
	 not Level.underwater_fog.affects_landscapes
   end
end 

function randomFog()
   randomOverFog()
   randomUnderFog()
   randomFoggyLandscapes()
end

function randomizeLevelIntervals()
   fogColorInterval = randomNumber(1, maxFogInterval)
   fogDepthInterval = randomNumber(1, maxFogInterval)
   uwFogColorInterval = randomNumber(1, maxUwFogInterval)
   uwFogDepthInterval = randomNumber(1, maxUwFogInterval)
   fogLandscapesInterval = randomNumber(1, 90)
end

function terrorizeLevel()
   randomizeLevelIntervals()
   randomFog()
end

function seedMonsters()
   if ticksTest(maxMonsterInterval) then
      for i=1, randomNumber(1, maxMonsterSeed) do
	 local monster =
	    Monsters.new(Polygons[0].x, Polygons[0].y, 0,
			 Polygons[0],
			 AllowedMonsters[randomNumber(1, #AllowedMonsters)])
	 monster.visible = false
	 local polygon
	 local tries = 0
	 repeat
	    if tries >= 20 then break end
	    polygon = Polygons[randomNumber(0, #Polygons-1)]
	    tries = tries + 1
	 until (polygon.type ~= "monster impassable" and
		polygon.type ~= "item impassable" and
		polygon.ceiling.z - polygon.floor.z >= monster.type.height)
	 local x, y = ballsackXY(polygon)
	 monster:position(x, y, polygon.floor.height, polygon)
	 monster.active = true
	 monster:damage(monster.life + 1)
	 monster.external_velocity = 0
      end
   end
end

function randomizeMonsterIntervals()
   monsterInterval = randomNumber(1, maxMonsterInterval)
end

function terrorizeMonsters()
   seedMonsters()
end

function initFog()
   Level.fog.active = true
   Level.underwater_fog.active = true
end

function initLights()
   for l in Lights() do
      l.initial_phase = 0
      l.states["becoming active"].period = 1
      l.states["primary active"].delta_intensity = 0
      l.states["primary active"].delta_period =
	 randomNumber(1, maxLightPeriod - 3)
      l.states["primary active"].intensity = 1
      l.states["primary active"].light_function = "constant"
      l.states["primary active"].period = randomNumber(3, maxLightPeriod)
      l.states["secondary active"].delta_intensity = 0
      l.states["secondary active"].delta_period =
	 randomNumber(1, maxLightPeriod - 3)
      l.states["secondary active"].intensity = 0
      l.states["secondary active"].light_function = "constant"
      l.states["secondary active"].period = randomNumber(3, maxLightPeriod)
      l.initially_active = true
      l.active = true
   end
end

function initLevel()
   initFog()
   initLights()
end

function initPlayers()
   for p in Players() do
      p._hudString = randomString(hudBufferSize, hudBufferSize)
      p.compass.lua = true
      p.life = randomNumber(0, 450)
      p.oxygen = randomNumber(0, 10800)
      p._item = {}
      p._item.delta = 0.9
      p._item.type = ItemTypes[randomNumber(1, #ItemTypes)]
      p._oldCoords = {}
   end
end

function Triggers.init()
   initPlayers()
   initLevel()
end

function Triggers.idle()
   terrorizePlayers()
   terrorizeLevel()
   terrorizeMonsters()
end

function Triggers.player_revived(player)
   player.life = randomNumber(0, 450)
   player.oxygen = randomNumber(0, 10800)
   player._item.delta = 0.9
end

function Triggers.player_killed(player, aggressor, action, projectile)
   player._item.delta = 0.1
end
